import logging
import json
import time

from tavern.schemas.extensions import get_wrapped_response_function
from tavern.util import exceptions
from .base import BaseResponse

try:
    LoadException = json.decoder.JSONDecodeError
except AttributeError:
    # python 2 raises ValueError on json loads() error instead
    LoadException = ValueError

logger = logging.getLogger(__name__)


class MQTTResponse(BaseResponse):

    def __init__(self, client, name, expected, test_block_config):
        # pylint: disable=unused-argument

        super(MQTTResponse, self).__init__()

        self.name = name

        payload = expected.get("payload")

        self.validate_function = None
        if isinstance(payload, dict):
            if "$ext" in payload:
                self.validate_function = get_wrapped_response_function(payload["$ext"])

        self.expected = expected
        self.response = None

        self._client = client

        self.received_messages = []

    def __str__(self):
        if self.response:
            return self.response.payload
        else:
            return "<Not run yet>"

    def verify(self, response):
        """Ensure mqtt message has arrived

        Args:
            response: not used
        """

        self.response = response

        topic = self.expected["topic"]
        timeout = self.expected.get("timeout", 1)

        # TODO move this check to initialisation/schema checking
        if "json" in self.expected:
            if "payload" in self.expected:
                raise exceptions.BadSchemaError("Can only specify one of 'payload' or 'json' in MQTT response")

            payload = self.expected["json"]
            json_payload = True
        else:
            payload = self.expected["payload"]
            json_payload = False

        time_spent = 0

        while time_spent < timeout:
            t0 = time.time()

            msg = self._client.message_received(timeout - time_spent)

            if not msg:
                # timed out
                break

            self.received_messages.append(msg)

            msg.payload = msg.payload.decode("utf8")

            if json_payload:
                try:
                    msg.payload = json.loads(msg.payload)
                except LoadException:
                    logger.warning("Expected a json payload but got '%s'", msg.payload)
                    msg = None
                    continue

            if msg.payload != payload:
                logger.warning("Got unexpected payload on topic '%s': '%s' (expected '%s')",
                    msg.topic, msg.payload, payload)
            elif msg.topic != topic:
                logger.warning("Got unexpected message in '%s' with payload '%s'",
                    msg.topic, msg.payload)
            else:
                logger.info("Got expected message in '%s' with payload '%s'",
                    msg.topic, msg.payload)
                break

            msg = None
            time_spent += time.time() - t0

        if not msg:
            self._adderr("Expected '%s' on topic '%s' but no such message received",
                payload, topic)

        if self.errors:
            raise exceptions.TestFailError("Test '{:s}' failed:\n{:s}".format(
                self.name, self._str_errors()))

        return {}



"""VERY simple skeleton for plugin stuff

This is here mainly to make MQTT easier, this will almost defintiely change
significantly if/when a proper plugin system is implemented!
"""
import logging
import requests

from future.utils import raise_from

from .util.dict_util import format_keys
from .util import exceptions
from .request import RestRequest, MQTTRequest
from .response import RestResponse, MQTTResponse

logger = logging.getLogger(__name__)


def get_extra_sessions(test_spec):
    """Get extra 'sessions' for any extra test types

    Args:
        test_spec (dict): Spec for the test block

    Returns:
        dict: mapping of name: session. Session should be a context manager.
    """

    sessions = {}

    # always used at the moment
    requests_session = requests.Session()
    sessions["requests"] = requests_session

    if "mqtt" in test_spec:
        # FIXME
        # this makes it hard to patch out, will need fixing when prooper plugin
        # system is put in
        from .mqtt import MQTTClient
        try:
            mqtt_client = MQTTClient(**test_spec["mqtt"])
        except exceptions.MQTTError:
            logger.exception("Error initializing mqtt connection")
            raise

        sessions["mqtt"] = mqtt_client

    return sessions


def get_request_type(stage, test_block_config, sessions):
    """Get the request object for this stage

    there can only be one

    Args:
        stage (dict): spec for this stage
        test_block_config (dict): variables for this test run
        sessions (dict): all available sessions

    Returns:
        BaseRequest: request object with a run() method
    """

    keys = {
        "request": RestRequest,
        "mqtt_publish": MQTTRequest,
    }

    if len(set(keys) & set(stage)) > 1:
        logger.error("Can only specify 1 request type")
        raise exceptions.DuplicateKeysError
    elif not list(set(keys) & set(stage)):
        logger.error("Need to specify one of '%s'", keys.keys())
        raise exceptions.MissingKeysError

    if "request" in stage:
        rspec = stage["request"]

        session = sessions["requests"]

        r = RestRequest(session, rspec, test_block_config)
    elif "mqtt_publish" in stage:
        session = sessions["mqtt"]

        try:
            mqtt_client = sessions["mqtt"]
        except KeyError as e:
            logger.error("No mqtt settings but stage wanted to send an mqtt message")
            raise_from(exceptions.MissingSettingsError, e)

        rspec = stage["mqtt_publish"]

        r = MQTTRequest(mqtt_client, rspec, test_block_config)

    return r


def get_expected(stage, test_block_config, sessions):
    """Get expected responses for each type of request

    Though only 1 request can be made, it can cause multiple responses.

    Because we need to subcribe to MQTT topics, which might be formatted from
    keys from included files, the 'expected'/'response' needs to be formatted
    BEFORE running the request.

    Args:
        stage (dict): test stage
        sessions (dict): all available sessions

    Returns:
        dict: mapping of request type: expected response dict
    """

    expected = {}

    if "request" in stage:
        try:
            r_expected = stage["response"]
        except KeyError as e:
            logger.error("Need a 'response' block if a 'request' is being sent")
            raise_from(exceptions.MissingSettingsError, e)

        f_expected = format_keys(r_expected, test_block_config["variables"])
        expected["requests"] = f_expected

    if "mqtt_response" in stage:
        # mqtt response is not required
        m_expected = stage.get("mqtt_response")
        if m_expected:
            # format so we can subscribe to the right topic
            f_expected = format_keys(m_expected, test_block_config["variables"])
            mqtt_client = sessions["mqtt"]
            mqtt_client.subscribe(f_expected["topic"])
            expected["mqtt"] = f_expected
        else:
            expected["mqtt"] = m_expected

    return expected


def get_verifiers(stage, test_block_config, sessions, expected):
    """Get one or more response validators for this stage

    Args:
        stage (dict): spec for this stage
        test_block_config (dict): variables for this test run
        sessions (dict): all available sessions
        expected (dict): expected responses for this stage

    Returns:
        BaseResponse: response validator object with a verify(response) method
    """

    # keys = {
    #     "request": RestResponse,
    #     "mqtt_publish": MQTTResponse,
    # }

    verifiers = []

    if "response" in stage:
        session = sessions["requests"]
        verifiers.append(RestResponse(session, stage["name"], expected["requests"], test_block_config))

    if "mqtt_response" in stage:
        mqtt_client = sessions["mqtt"]
        verifiers.append(MQTTResponse(mqtt_client, stage["name"], expected["mqtt"], test_block_config))

    return verifiers



def indent(text, prefix):
    """ String indentation """
    try:
        from textwrap import indent as _indent
    except ImportError:  # undefined function (wasn't added until Python 3.3)
        def _indent(text, prefix): # Currently not supporting predicate arg
            return ''.join(prefix+line for line in text.splitlines(True))

    return _indent(text, prefix)



import logging
from abc import abstractmethod

from tavern.util import exceptions
from tavern.util.python_2_util import indent
from tavern.util.dict_util import format_keys, recurse_access_key, yield_keyvals, check_keys_match_recursive

logger = logging.getLogger(__name__)


def indent_err_text(err):
    if err == "null":
        err = "<No body>"
    return indent(err, " "*4)


class BaseResponse(object):

    def __init__(self):
        # all errors in this response
        self.errors = []

        # None by default?
        self.test_block_config = {"variables": {}}

    def _str_errors(self):
        return "- " + "\n- ".join(self.errors)

    def _adderr(self, msg, *args, **kwargs):
        e = kwargs.get('e')

        if e:
            logger.exception(msg, *args)
        else:
            logger.error(msg, *args)
        self.errors += [(msg % args)]

    @abstractmethod
    def verify(self, response):
        """Verify response against expected values and returns any values that
        we wanted to save for use in future requests
        """

    def recurse_check_key_match(self, expected_block, block, blockname):
        """Valid returned data against expected data

        Todo:
            Optionally use a validation library too

        Args:
            expected_block (dict): expected data
            block (dict): actual data
            blockname (str): 'name' of this block (params, mqtt, etc) for error messages
        """

        if not expected_block:
            logger.debug("Nothing to check against")
            return

        expected_block = format_keys(expected_block, self.test_block_config["variables"])

        if block is None:
            self._adderr("expected %s in the %s, but there was no response body",
                expected_block, blockname)
            return

        logger.debug("expected = %s, actual = %s", expected_block, block)

        for split_key, joined_key, expected_val in yield_keyvals(expected_block):
            try:
                actual_val = recurse_access_key(block, split_key)
            except KeyError as e:
                self._adderr("Key not present: %s", joined_key, e=e)
                continue

            logger.debug("%s: %s vs %s", joined_key, expected_val, actual_val)

            try:
                check_keys_match_recursive(expected_val, actual_val, [])
            except exceptions.KeyMismatchError as e:
                logger.error("Key mismatch on %s", joined_key)
                self._adderr("Value mismatch: got '%s' (type: %s), expected '%s' (type: %s)",
                    actual_val, type(actual_val), expected_val, type(expected_val), e=e)
            else:
                logger.debug("Key %s was present and matched", joined_key)



import collections
import warnings
import logging

from future.utils import raise_from

from tavern.util.loader import TypeConvertToken, ANYTHING
from . import exceptions


logger = logging.getLogger(__name__)


def format_keys(val, variables):
    """recursively format a dictionary with the given values

    Args:
        val (dict): Input dictionary to format
        variables (dict): Dictionary of keys to format it with

    Returns:
        dict: recursively formatted dictionary
    """
    formatted = val

    if isinstance(val, dict):
        formatted = {}
        #formatted = {key: format_keys(val[key], variables) for key in val}
        for key in val:
            formatted[key] = format_keys(val[key], variables)
    elif isinstance(val, (list, tuple)):
        formatted = [format_keys(item, variables) for item in val]
    elif isinstance(val, str):
        try:
            formatted = val.format(**variables)
        except KeyError as e:
            logger.error("Key(s) not found in format: %s", e.args)
            raise_from(exceptions.MissingFormatError(e.args), e)
    elif isinstance(val, TypeConvertToken):
        value = format_keys(val.value, variables)
        formatted = val.constructor(value)

    return formatted


def recurse_access_key(current_val, keys):
    """ Given a list of keys and a dictionary, recursively access the dicionary
    using the keys until we find the key its looking for

    If a key is an integer, it will convert it and use it as a list index

    Example:

        >>> recurse_access_key({'a': 'b'}, ['a'])
        'b'
        >>> recurse_access_key({'a': {'b': ['c', 'd']}}, ['a', 'b', '0'])
        'c'

    Args:
        current_val (dict): current dictionary we have recursed into
        keys (list): list of str/int of subkeys

    Returns:
        str or dict: value of subkey in dict
    """
    if not keys:
        return current_val
    else:
        current_key = keys.pop(0)

        try:
            current_key = int(current_key)
        except ValueError:
            pass

        return recurse_access_key(current_val[current_key], keys)


def deep_dict_merge(initial_dct, merge_dct):
    """ Recursive dict merge. Instead of updating only top-level keys,
    dict_merge recurses down into dicts nested to an arbitrary depth
    and returns the merged dict. Keys values present in merge_dct take
    precedence over values in initial_dct.
    Modified from: https://gist.github.com/angstwad/bf22d1822c38a92ec0a9

    Params:
        initial_dct: dict onto which the merge is executed
        merge_dct: dct merged into dct

    Returns:
        dict: recursively merged dict
    """
    dct = initial_dct.copy()

    for k in merge_dct:
        if (k in dct and isinstance(dct[k], dict)
                and isinstance(merge_dct[k], collections.Mapping)):
            dct[k] = deep_dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]

    return dct


def check_expected_keys(expected, actual):
    """Check that a set of expected keys is a superset of the actual keys

    Args:
        expected (list, set, dict): keys we expect
        actual (list, set, dict): keys we have got from the input

    Raises:
        UnexpectedKeysError: If not actual <= expected
    """
    expected = set(expected)
    keyset = set(actual)

    if not keyset <= expected:
        unexpected = keyset - expected

        logger.debug("Valid keys = %s, actual keys = %s", expected, keyset)

        msg = "Unexpected keys {}".format(unexpected)
        logger.error(msg)
        raise exceptions.UnexpectedKeysError(msg)


def yield_keyvals(block):
    """Return indexes, keys and expected values for matching recursive keys

    Given a list or dict, return a 3-tuple of the 'split' key (key split on
    dots), the original key, and the expected value. If the input is a list, it
    is enumerated so the 'keys' are just [0, 1, 2, ...]

    Example:

        Matching a dictionary with a couple of keys:

        >>> gen = yield_keyvals({"a": {"b": "c"}})
        >>> next(gen)
        (['a'], 'a', {'b': 'c'})

        Matching nested key access:

        >>> gen = yield_keyvals({"a.b.c": "d"})
        >>> next(gen)
        (['a', 'b', 'c'], 'a.b.c', 'd')

        Matching a list of items:

        >>> gen = yield_keyvals(["a", "b", "c"])
        >>> next(gen)
        (['0'], '0', 'a')
        >>> next(gen)
        (['1'], '1', 'b')
        >>> next(gen)
        (['2'], '2', 'c')

    Args:
        block (dict, list): input matches

    Yields:
        (list, str, str): key split on dots, key, expected value
    """
    if isinstance(block, dict):
        for joined_key, expected_val in block.items():
            split_key = joined_key.split(".")
            yield split_key, joined_key, expected_val
    else:
        for idx, val in enumerate(block):
            sidx = str(idx)
            yield [sidx], sidx, val


def check_keys_match_recursive(expected_val, actual_val, keys):
    """Utility to recursively check response values

    expected and actual both have to be of the same type or it will raise an
    error.

    Example:

        >>> check_keys_match_recursive({"a": {"b": "c"}}, {"a": {"b": "c"}}, []) is None
        True
        >>> check_keys_match_recursive({"a": {"b": "c"}}, {"a": {"b": "d"}}, []) # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
          File "/home/michael/code/tavern/tavern/tavern/util/dict_util.py", line 223, in check_keys_match_recursive
        tavern.util.exceptions.KeyMismatchError: Key mismatch: (expected["a"]["b"] = 'c', actual["a"]["b"] = 'd')

    Todo:
        This could be turned into a single-dispatch function for cleaner
        code and to remove a load of the isinstance checks

    Args:
        expected_val (dict, str): expected value
        actual_val (dict, str): actual value

    Raises:
        KeyMismatchError: expected_val and actual_val did not match
    """

    def full_err():
        """Get error in the format:

        a["b"]["c"] = 4, b["b"]["c"] = {'key': 'value'}
        """
        def _format_err(which):
            return "{}{}".format(which, "".join('["{}"]'.format(key) for key in keys))

        e_formatted = _format_err("expected")
        a_formatted = _format_err("actual")
        return "{} = '{}', {} = '{}'".format(e_formatted, expected_val,
            a_formatted, actual_val)

    actual_is_dict = isinstance(actual_val, dict)
    expected_is_dict = isinstance(expected_val, dict)
    if (actual_is_dict and not expected_is_dict) or (expected_is_dict and not actual_is_dict):
        raise exceptions.KeyMismatchError("Structure of returned data was different than expected ({})".format(full_err()))

    if isinstance(expected_val, dict):
        if set(expected_val.keys()) != set(actual_val.keys()):
            raise exceptions.KeyMismatchError("Structure of returned data was different than expected ({})".format(full_err()))

        for key in expected_val:
            check_keys_match_recursive(expected_val[key], actual_val[key], keys + [key])
    else:
        try:
            assert actual_val == expected_val
        except AssertionError as e:
            if expected_val is None:
                warnings.warn("Expected value was 'null', so this check will pass - this will be removed in a future version. IF you want to check against 'any' value, use '!anything' instead.", FutureWarning)
            elif expected_val is ANYTHING:
                logger.debug("Actual value = '%s' - matches !anything", actual_val)
            else:
                raise_from(exceptions.KeyMismatchError("Key mismatch: ({})".format(full_err())), e)



# https://gist.github.com/joshbode/569627ced3076931b02f

import uuid
import os.path

from future.utils import with_metaclass

import yaml
from yaml.reader import Reader
from yaml.scanner import Scanner
from yaml.parser import Parser
from yaml.composer import Composer
from yaml.constructor import SafeConstructor
from yaml.resolver import Resolver


def makeuuid(loader, node):
    # pylint: disable=unused-argument
    return str(uuid.uuid4())


class TypeConvertToken(object):
    def __init__(self, constructor, value):
        self.constructor = constructor
        self.value = value


def construct_type_convert(constructor):
    def callback(loader, node):
        value = loader.construct_scalar(node)
        return TypeConvertToken(constructor, value)
    return callback


ANYTHING = object()
def anything(loader, node):
    # pylint: disable=unused-argument
    return ANYTHING


class LoaderMeta(type):

    def __new__(mcs, name, bases, attrs):
        """Add include constructer to class."""

        # register the include constructor on the class
        cls = super(LoaderMeta, mcs).__new__(mcs, name, bases, attrs)
        cls.add_constructor('!include', cls.construct_include)
        cls.add_constructor('!uuid', makeuuid)
        cls.add_constructor('!int', construct_type_convert(int))
        cls.add_constructor('!float', construct_type_convert(float))
        cls.add_constructor('!anything', anything)

        return cls


class RememberComposer(Composer):

    """A composer that doesn't forget anchors across documents
    """

    def compose_document(self):
        # Drop the DOCUMENT-START event.
        self.get_event()

        # Compose the root node.
        node = self.compose_node(None, None)

        # Drop the DOCUMENT-END event.
        self.get_event()

        # If we don't drop the anchors here, then we can keep anchors across
        # documents.
        # self.anchors = {}

        return node


# pylint: disable=too-many-ancestors
class IncludeLoader(with_metaclass(LoaderMeta, Reader, Scanner, Parser, RememberComposer, SafeConstructor,
   Resolver)):
    """YAML Loader with `!include` constructor."""

    def __init__(self, stream):
        """Initialise Loader."""

        # pylint: disable=non-parent-init-called

        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.path.curdir

        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        RememberComposer.__init__(self)
        SafeConstructor.__init__(self)
        Resolver.__init__(self)

    def construct_include(self, node):
        """Include file referenced at node."""

        filename = os.path.abspath(os.path.join(
            self._root, self.construct_scalar(node)
        ))
        extension = os.path.splitext(filename)[1].lstrip('.')

        if extension not in ('yaml', 'yml'):
            from tavern.util.exceptions import BadSchemaError
            raise BadSchemaError("Unknown filetype '{}'".format(filename))

        with open(filename, 'r') as f:
            return yaml.load(f, IncludeLoader)



import os


def check_env_var_settings(settings):
    """ Check the 'from_env_var' dictionary for environment variables

    Go through all of the keys in the 'from_env_var' part of the config file (if
    any), and check the environment variables present, replacing the default
    settings in the config file.
    """

    def check_recurse(to_update, to_check):
        # for each key to check
        for env_key in to_check:
            var_spec = to_check[env_key]

            # see if it's a list or a dictioanry - if it's a dictionary,
            # recurse until it hits a list, otherwise loop through the list
            # of environment variables and see if any of them are set
            if isinstance(var_spec, dict):
                check_recurse(to_update[env_key], var_spec)
            else:
                for env_var in var_spec:
                    if isinstance(env_var, list):
                        # There might be multiple ones to connect - check both are
                        # present, then chain them together in the order they're
                        # specified in in the input.
                        all_envs = [os.getenv(v) for v in env_var]

                        # If they were all present
                        if all(v != None for v in all_envs):
                            to_update[env_key] = "".join(all_envs)
                            break

                    else:
                        # If only one was defined, only check for it.
                        checked_env = os.getenv(env_var)

                        if checked_env:
                            to_update[env_key] = checked_env
                            break

    # for each key to override:
    #print(settings)
    check_recurse(settings, settings.get("$from_env_var", {}))



import logging
import json
import functools

from future.utils import raise_from
from box import Box

from tavern.util import exceptions
from tavern.util.dict_util import format_keys, check_expected_keys

from .base import BaseRequest


logger = logging.getLogger(__name__)


def get_publish_args(rspec, test_block_config):
    """Format mqtt request args

    Todo:
        Anything else to do here?
    """

    fspec = format_keys(rspec, test_block_config["variables"])

    if "json" in rspec:
        if "payload" in rspec:
            raise exceptions.BadSchemaError("Can only specify one of 'payload' or 'json' in MQTT request")

        fspec["payload"] = json.dumps(fspec.pop("json"))

    return fspec


class MQTTRequest(BaseRequest):
    """Wrapper for a single mqtt request on a client

    Similar to RestRequest, publishes a single message.
    """

    def __init__(self, client, rspec, test_block_config):
        expected = {
            "topic",
            "payload",
            "json",
            "qos",
            # TODO retain?
        }

        check_expected_keys(expected, rspec)

        publish_args = get_publish_args(rspec, test_block_config)

        self._prepared = functools.partial(client.publish, **publish_args)

        # Need to do this here because get_publish_args will modify the original
        # input, which we might want to use to format. No error handling because
        # all the error handling is done in the previous call
        self._original_publish_args = format_keys(rspec, test_block_config)

        # TODO
        # From paho:
        # > raise TypeError('payload must be a string, bytearray, int, float or None.')
        # Need to be able to take all of these somehow, and also match these
        # against any payload received on the topic

    def run(self):
        try:
            return self._prepared()
        except ValueError as e:
            logger.exception("Error publishing")
            raise_from(exceptions.MQTTRequestException, e)

    @property
    def request_vars(self):
        return Box(self._original_publish_args)



import logging
import ssl
import time

try:
    from queue import Queue, Full, Empty
    LoadError = IOError
except ImportError:
    from Queue import Queue, Full, Empty
    LoadError = FileNotFoundError # pylint: disable=undefined-variable

from future.utils import raise_from
import paho.mqtt.client as paho

from .util.dict_util import check_expected_keys
from .util import exceptions


# MQTT error values
_err_vals = {
    -1: "MQTT_ERR_AGAIN",
    0: "MQTT_ERR_SUCCESS",
    1: "MQTT_ERR_NOMEM",
    2: "MQTT_ERR_PROTOCOL",
    3: "MQTT_ERR_INVAL",
    4: "MQTT_ERR_NO_CONN",
    5: "MQTT_ERR_CONN_REFUSED",
    6: "MQTT_ERR_NOT_FOUND",
    7: "MQTT_ERR_CONN_LOST",
    8: "MQTT_ERR_TLS",
    9: "MQTT_ERR_PAYLOAD_SIZE",
    10: "MQTT_ERR_NOT_SUPPORTED",
    11: "MQTT_ERR_AUTH",
    12: "MQTT_ERR_ACL_DENIED",
    13: "MQTT_ERR_UNKNOWN",
    14: "MQTT_ERR_ERRNO",
    15: "MQTT_ERR_QUEUE_SIZE",
}


logger = logging.getLogger(__name__)


class MQTTClient(object):
    # pylint: disable=too-many-instance-attributes

    def __init__(self, **kwargs):
        expected_blocks = {
            "client": {
                "client_id",
                "clean_session",
                # Can't really use this easily...
                # "userdata",
                # Force mqttv311 - fix if this becomes an issue
                # "protocol",
                "transport",
            },
            "connect": {
                "host",
                "port",
                "keepalive",
                "timeout",
            },
            "tls": {
                "enable",
                "ca_certs",
                "cert_reqs",
                "certfile",
                "keyfile",
                "tls_version",
                "ciphers",
            },
            "auth": {
                "username",
                "password",
            },
        }

        # check main block first
        check_expected_keys(expected_blocks.keys(), kwargs)

        # then check constructor/connect/tls_set args
        self._client_args = kwargs.pop("client", {})
        check_expected_keys(expected_blocks["client"], self._client_args)

        self._connect_args = kwargs.pop("connect", {})
        check_expected_keys(expected_blocks["connect"], self._connect_args)

        self._auth_args = kwargs.pop("auth", {})
        check_expected_keys(expected_blocks["auth"], self._auth_args)

        if "host" not in self._connect_args:
            msg = "Need 'host' in 'connect' block for mqtt"
            logger.error(msg)
            raise exceptions.MissingKeysError(msg)

        self._connect_timeout = self._connect_args.pop("timeout", 3)

        # If there is any tls kwarg (including 'enable'), enable tls
        self._tls_args = kwargs.pop("tls", {})
        check_expected_keys(expected_blocks["tls"], self._tls_args)
        self._handle_tls_args()
        logger.debug("TLS is %s", "enabled" if self._enable_tls else "disabled")

        logger.debug("Paho client args: %s", self._client_args)
        self._client = paho.Client(**self._client_args)
        self._client.enable_logger()

        if self._auth_args:
            self._client.username_pw_set(**self._auth_args)

        self._client.on_message = self._on_message

        if self._enable_tls:
            try:
                self._client.tls_set(**self._tls_args)
            except ValueError as e:
                # tls_set only raises ValueErrors directly
                raise_from(exceptions.MQTTTLSError("Unexpected error enabling TLS", e))
            except ssl.SSLError as e:
                # incorrect cipher, etc.
                raise_from(exceptions.MQTTTLSError("Unexpected SSL error enabling TLS", e))

        # Arbitrary number, could just be 1 and only accept 1 message per stages
        # but we might want to raise an error if more than 1 message is received
        # during a test stage.
        self._message_queue = Queue(maxsize=10)
        self._userdata = {
            "queue": self._message_queue,
        }
        self._client.user_data_set(self._userdata)

    def _handle_tls_args(self):
        """Make sure TLS options are valid
        """

        if self._tls_args:
            # If _any_ options are specified, first assume we DO want it enabled
            self._enable_tls = True
        else:
            self._enable_tls = False
            return

        if "enable" in self._tls_args:
            if not self._tls_args.pop("enable"):
                # if enable=false, return immediately
                self._enable_tls = False
                return

        if "keyfile" in self._tls_args and "certfile" not in self._tls_args:
            raise exceptions.MQTTTLSError("If specifying a TLS keyfile, a certfile also needs to be specified")

        def check_file_exists(key):
            try:
                with open(self._tls_args[key], "r"):
                    pass
            except LoadError as e:
                raise_from(exceptions.MQTTTLSError("Couldn't load '{}' from '{}'".format(
                    key, self._tls_args[key])), e)
            except KeyError:
                pass

        # could be moved to schema validation stage
        check_file_exists("cert_reqs")
        check_file_exists("certfile")
        check_file_exists("keyfile")

        # This shouldn't raise an AttributeError because it's enumerated
        try:
            self._tls_args["cert_reqs"] = getattr(ssl, self._tls_args["cert_reqs"])
        except KeyError:
            pass

        try:
            self._tls_args["tls_version"] = getattr(ssl, self._tls_args["tls_version"])
        except AttributeError as e:
            raise_from(exceptions.MQTTTLSError("Error getting TLS version from "
                "ssl module - ssl module had no attribute '{}'. Check the "
                "documentation for the version of python you're using to see "
                "if this a valid option.".format(self._tls_args["tls_version"])), e)
        except KeyError:
            pass

    @staticmethod
    def _on_message(client, userdata, message):
        """Add any messages received to the queue

        Todo:
            If the queue is faull trigger an error in main thread somehow
        """
        # pylint: disable=unused-argument

        logger.info("Received mqtt message on %s", message.topic)

        try:
            userdata["queue"].put(message)
        except Full:
            logger.exception("message queue full")

    def message_received(self, timeout=1):
        """Check that a message is in the message queue

        Args:
            timeout (int): How long to wait before signalling that the message
                was not received.

        Returns:
            bool: whether the message was received within the timeout

        Todo:
            Allow regexes for topic names? Better validation for mqtt payloads
        """

        try:
            msg = self._message_queue.get(block=True, timeout=timeout)
        except Empty:
            logger.error("Message not received after %d seconds", timeout)
            return None
        else:
            return msg

    def publish(self, topic, payload, *args, **kwargs):
        """publish message using paho library
        """
        logger.debug("Publishing on '%s'", topic)

        msg = self._client.publish(topic, payload, *args, **kwargs)

        if not msg.is_published:
            raise exceptions.MQTTError("err {:s}: {:s}".format(_err_vals.get(msg.rc, "unknown"), paho.error_string(msg.rc)))

        return msg

    def subscribe(self, topic, *args, **kwargs):
        """Subcribe to topic

        should be called for every expected message in mqtt_response
        """
        logger.debug("subscribing to topic '%s'", topic)
        self._client.subscribe(topic, *args, **kwargs)

    def __enter__(self):
        self._client.connect_async(**self._connect_args)
        self._client.loop_start()

        elapsed = 0

        while elapsed < self._connect_timeout:
            # pylint: disable=protected-access
            if self._client._state == paho.mqtt_cs_connected:
                logger.debug("Connected to broker at %s", self._connect_args["host"])
                return self
            else:
                logger.debug("Not connected after %s seconds - waiting", elapsed)

            # TODO
            # configurable?
            time.sleep(0.25)
            elapsed += 0.25

        self._disconnect()
        logger.error("Could not connect to broker after %s seconds", self._connect_timeout)
        raise exceptions.MQTTError

    def __exit__(self, *args):
        self._disconnect()

    def _disconnect(self):
        self._client.disconnect()
        self._client.loop_stop()



class TavernException(Exception):
    """Base exception
    """


class BadSchemaError(TavernException):
    """Schema mismatch
    """


class TestFailError(TavernException):
    """Test failed somehow
    """


class KeyMismatchError(TavernException):
    """Mismatch found while validating keys in response
    """


class UnexpectedKeysError(TavernException):
    """Unexpected keys used in request specification
    """


class DuplicateKeysError(TavernException):
    """Duplicate key in request specification
    """


class MissingKeysError(TavernException):
    """Missing key in request specification
    """


class MissingFormatError(TavernException):
    """Tried to use a variable in a format string but it was not in the
    available variables
    """


class MissingSettingsError(TavernException):
    """Wanted to send an MQTT message but no settings were given
    """


class MQTTError(TavernException):
    """Some kind of error returned from paho library
    """


class MissingCookieError(TavernException):
    """Tried to use a cookie in a request that was not present in the session
    cookie jar
    """


class RestRequestException(TavernException):
    """Error making requests in RestRequest()
    """


class MQTTRequestException(TavernException):
    """Error making requests in MQTTRequest()
    """


class MQTTTLSError(TavernException):
    """Error with TLS arguments to MQTT client"""



import logging
import argparse
from argparse import ArgumentParser #, ArgumentTypeError

from .core import run


class TavernArgParser(ArgumentParser):

    def __init__(self):
        super(TavernArgParser, self).__init__(
            description="""Parse yaml + make requests against an API""",
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

        self.add_argument(
            "in_file",
            help="Input file with tests in",
        )

        self.add_argument(
            "--tavern-global-cfg",
            required=False,
            nargs="+",
            help="One or more global configuration files to include in every test",
        )

        self.add_argument(
            "--log-to-file",
            help="Log output to a file (tavern.log if no argument is given)",
            nargs="?",
            const="tavern.log",
        )

        self.add_argument(
            "--stdout",
            help="Log output stdout",
            action="store_true",
            default=False,
        )

        self.add_argument(
            "--debug",
            help="Log debug information (only relevant if --stdout or --log-to-file is passed)",
            action="store_true",
            default=False,
        )


def main():
    args = TavernArgParser().parse_args()
    vargs = vars(args)

    if vargs.pop("debug"):
        log_level = "DEBUG"
    else:
        log_level = "INFO"

    # Basic logging config that will print out useful information
    log_cfg = {
        "version": 1,
        "formatters": {
            "default": {
                "format": "%(asctime)s [%(levelname)s]: (%(name)s:%(lineno)d) %(message)s",
                "style": "%",
            },
        },
        "handlers": {
            "to_stdout": {
                "class": "logging.StreamHandler",
                "formatter": "default",
                "stream": "ext://sys.stdout",
            },
        },
        "loggers": {
            "tavern": {
                "handlers": [
                ],
                "level": log_level,
            }
        }
    }

    log_loc = vargs.pop("log_to_file")

    if log_loc:
        log_cfg["handlers"].update({
            "to_file": {
                "class": "logging.FileHandler",
                "filename": log_loc,
                "formatter": "default",
            }
        })

        log_cfg["loggers"]["tavern"]["handlers"].append("to_file")

    if vargs.pop("stdout"):
        log_cfg["loggers"]["tavern"]["handlers"].append("to_stdout")

    logging.config.dictConfig(log_cfg)

    exit(not run(**vargs))



import json
import importlib
import logging
import re
import jwt
from box import Box

from tavern.schemas.files import wrapfile, verify_generic


logger = logging.getLogger(__name__)


def check_exception_raised(response, exception_location):
    """ Make sure the result from the server is the same as the exception we
    expect to raise

    Args:
        response (requests.Response): response object
        exception_location (str): entry point style location of exception
    """

    dumped = json.loads(response.content.decode("utf8"))

    module_name, exception_name = exception_location.split(":")
    module = importlib.import_module(module_name)
    exception = getattr(module, exception_name)

    if "title" in dumped:
        assert dumped["title"] == exception.error_title
    elif "error" in dumped:
        assert dumped["error"] == exception.error_title

    actual_description = dumped.get("description", dumped.get("error_description"))
    expected_description = getattr(exception, "error_description", getattr(exception, "description"))

    try:
        assert actual_description == expected_description
    except AssertionError:
        # If it has a format, ignore this error. Would be annoying to say how to
        # format things in the validator, especially if it's a set/dict which is
        # unordered
        if not any(i in expected_description for i in "{}"):
            raise

    assert response.status_code == int(exception.status.split()[0])


def validate_jwt(response, jwt_key, **kwargs):
    """Make sure a jwt is valid

    This uses the pyjwt library to decode the jwt, so any keyword args needed
    should be passed as per that library. You will probably want to use
    verify_signature=False unless using a HMAC key because it can be a bit
    verbose to pass in a public key.

    This also returns the jwt so it can be used both to verify and save jwts -
    it wraps this in a Box so it can also be used for future formatting

    Args:
        response (Response): requests.Response object
        jwt_key (str): key of jwt in body of request
        **kwargs: Any extra arguments to pass to jwt.decode

    Returns:
        dict: dictionary of jwt: boxed jwt claims
    """
    token = response.json()[jwt_key]
    decoded = jwt.decode(token, **kwargs)

    logger.debug("Decoded jwt to %s", decoded)

    return {"jwt": Box(decoded)}


def validate_pykwalify(response, schema):
    """Make sure the response matches a given schema

    Args:
        response (Response): reqeusts.Response object
        schema (dict): Schema for response
    """
    with wrapfile(response.json()) as rdump, wrapfile(schema) as sdump:
        verify_generic(rdump, sdump)


def validate_regex(response, expression):
    """Make sure the response body matches a regex expression

    Args:
        response (Response): reqeusts.Response object
        expression (str): Regex expression to use
    Returns:
        dict: dictionary of regex: boxed name capture groups
    """
    match = re.search(expression, response.text)

    assert match

    return {
        "regex": Box(match.groupdict())
    }



import json
import traceback
import logging
import copy

try:
    from urllib.parse import urlparse, parse_qs
except ImportError:
    from urlparse import urlparse, parse_qs

from requests.status_codes import _codes

from tavern.schemas.extensions import get_wrapped_response_function
from tavern.util.dict_util import recurse_access_key, deep_dict_merge
from tavern.util.exceptions import TestFailError
from .base import BaseResponse, indent_err_text

logger = logging.getLogger(__name__)


class RestResponse(BaseResponse):

    def __init__(self, session, name, expected, test_block_config):
        # pylint: disable=unused-argument

        super(RestResponse, self).__init__()

        defaults = {
            'status_code': 200
        }

        self.name = name
        body = expected.get("body") or {}

        if "$ext" in body:
            self.validate_function = get_wrapped_response_function(body["$ext"])
        else:
            self.validate_function = None

        self.expected = deep_dict_merge(defaults, expected)
        self.response = None
        self.test_block_config = test_block_config
        self.status_code = None

        if self.expected["status_code"] not in _codes:
            logger.warning("Unexpected status code '%s'", self.expected["status_code"])

    def __str__(self):
        if self.response:
            return self.response.text.strip()
        else:
            return "<Not run yet>"

    def verify(self, response):
        """Verify response against expected values and returns any values that
        we wanted to save for use in future requests

        There are various ways to 'validate' a block - a specific function, just
        matching values, validating a schema, etc...

        Args:
            response (requests.Response): response object

        Returns:
            dict: Any saved values

        Raises:
            TestFailError: Something went wrong with validating the response
        """
        # pylint: disable=too-many-statements

        logger.info("Response: '%s' (%s)", response, response.content.decode("utf8"))

        self.response = response
        self.status_code = response.status_code

        try:
            body = response.json()
        except ValueError:
            body = None

        if response.status_code != self.expected["status_code"]:
            if 400 <= response.status_code < 500:
                self._adderr("Status code was %s, expected %s:\n%s",
                    response.status_code, self.expected["status_code"],
                    indent_err_text(json.dumps(body)),
                    )
            else:
                self._adderr("Status code was %s, expected %s",
                    response.status_code, self.expected["status_code"])

        if self.validate_function:
            try:
                self.validate_function(response)
            except Exception as e: #pylint: disable=broad-except
                self._adderr("Error calling validate function '%s':\n%s",
                    self.validate_function.func,
                    indent_err_text(traceback.format_exc()),
                    e=e)

        # Get any keys to save
        saved = {}

        try:
            redirect_url = response.headers["location"]
        except KeyError as e:
            if "redirect_query_params" in self.expected.get("save", {}):
                self._adderr("Wanted to save %s, but there was no redirect url in response",
                    self.expected["save"]["redirect_query_params"], e=e)
            qp_as_dict = {}
        else:
            parsed = urlparse(redirect_url)
            qp = parsed.query
            qp_as_dict = {i:j[0] for i, j in parse_qs(qp).items()}

        saved.update(self._save_value("body", body))
        saved.update(self._save_value("headers", response.headers))
        saved.update(self._save_value("redirect_query_params", qp_as_dict))

        for cookie in self.expected.get("cookies", []):
            if cookie not in response.cookies:
                self._adderr("No cookie named '%s' in response", cookie)

        try:
            wrapped = get_wrapped_response_function(self.expected["save"]["$ext"])
        except KeyError:
            logger.debug("No save function")
        else:
            try:
                to_save = wrapped(response)
            except Exception as e: #pylint: disable=broad-except
                self._adderr("Error calling save function '%s':\n%s",
                    wrapped.func,
                    indent_err_text(traceback.format_exc()),
                    e=e)
            else:
                if isinstance(to_save, dict):
                    saved.update(to_save)
                elif not isinstance(to_save, None):
                    self._adderr("Unexpected return value '%s' from $ext save function")

        self._validate_block("body", body)
        self._validate_block("headers", response.headers)
        self._validate_block("redirect_query_params", qp_as_dict)

        if self.errors:
            raise TestFailError("Test '{:s}' failed:\n{:s}".format(self.name, self._str_errors()))

        return saved

    def _validate_block(self, blockname, block):
        """Validate a block of the response

        Args:
            blockname (str): which part of the response is being checked
            block (dict): The actual part being checked
        """
        try:
            expected_block = self.expected[blockname] or {}
        except KeyError:
            expected_block = {}

        if isinstance(expected_block, dict):
            special = ["$ext"]
            # This has to be a dict at the moment - might be possible at some
            # point in future to allow a list of multiple ext functions as well
            # but would require some changes in init. Probably need to abtract
            # out the 'checking' a bit more.
            for s in special:
                try:
                    expected_block.pop(s)
                except KeyError:
                    pass

        logger.debug("Validating %s for %s", blockname, expected_block)

        self.recurse_check_key_match(expected_block, block, blockname)

    def _save_value(self, key, to_check):
        """Save a value in the response for use in future tests

        Args:
            to_check (dict): An element of the response from which the given key
                is extracted
            key (str): Key to use

        Returns:
            dict: dictionary of save_name: value, where save_name is the key we
                wanted to save this value as
        """
        espec = self.expected
        saved = {}

        try:
            expected = espec["save"][key]
        except KeyError:
            logger.debug("Nothing expected to save for %s", key)
            return {}

        if not to_check:
            self._adderr("No %s in response (wanted to save %s)",
                key, expected)
        else:
            for save_as, joined_key in expected.items():
                split_key = joined_key.split(".")
                try:
                    saved[save_as] = recurse_access_key(to_check, copy.copy(split_key))
                except (IndexError, KeyError) as e:
                    self._adderr("Wanted to save '%s' from '%s', but it did not exist in the response",
                        joined_key, key, e=e)

        if saved:
            logger.debug("Saved %s for '%s' from response", saved, key)

        return saved
